EXPORTS
	alice_signal_protocol_parameters_create
	alice_signal_protocol_parameters_destroy
	bob_signal_protocol_parameters_create
	bob_signal_protocol_parameters_destroy
	ciphertext_message_get_serialized
	ciphertext_message_get_type
	curve_calculate_agreement
	curve_calculate_signature
	curve_calculate_vrf_signature
	curve_decode_point
	curve_decode_point_ed
	curve_decode_point_mont
	curve_decode_private_point
	curve_generate_key_pair
	curve_generate_private_key
	curve_generate_public_key
	curve_internal_fast_tests
	curve_verify_signature
	curve_verify_vrf_signature
	ec_key_pair_create
	ec_key_pair_destroy
	ec_key_pair_get_private
	ec_key_pair_get_public
	ec_private_key_compare
	ec_private_key_destroy
	ec_private_key_serialize
	ec_public_key_compare
	ec_public_key_destroy
	ec_public_key_get_ed
	ec_public_key_get_mont
	ec_public_key_list_alloc
	ec_public_key_list_at
	ec_public_key_list_copy
	ec_public_key_list_free
	ec_public_key_list_push_back
	ec_public_key_list_size
	ec_public_key_list_sort
	ec_public_key_memcmp
	ec_public_key_serialize
	ec_public_key_serialize_omemo
	hkdf_compare
	hkdf_create
	hkdf_derive_secrets
	hkdf_destroy
	pre_key_signal_message_copy
	pre_key_signal_message_create
	pre_key_signal_message_deserialize
	pre_key_signal_message_deserialize_omemo
	pre_key_signal_message_destroy
	pre_key_signal_message_get_base_key
	pre_key_signal_message_get_identity_key
	pre_key_signal_message_get_message_version
	pre_key_signal_message_get_pre_key_id
	pre_key_signal_message_get_registration_id
	pre_key_signal_message_get_signal_message
	pre_key_signal_message_get_signed_pre_key_id
	pre_key_signal_message_has_pre_key_id
	ratchet_chain_key_create
	ratchet_chain_key_create_next
	ratchet_chain_key_destroy
	ratchet_chain_key_get_index
	ratchet_chain_key_get_key
	ratchet_chain_key_get_message_keys
	ratchet_identity_key_pair_create
	ratchet_identity_key_pair_deserialize
	ratchet_identity_key_pair_destroy
	ratchet_identity_key_pair_get_private
	ratchet_identity_key_pair_get_public
	ratchet_identity_key_pair_serialize
	ratcheting_session_alice_initialize
	ratcheting_session_bob_initialize
	ratcheting_session_symmetric_initialize
	ratchet_kdf_infos_for_session_version
	ratchet_root_key_compare
	ratchet_root_key_create
	ratchet_root_key_create_chain
	ratchet_root_key_destroy
	ratchet_root_key_get_key
	sender_chain_key_create
	sender_chain_key_create_message_key
	sender_chain_key_create_next
	sender_chain_key_destroy
	sender_chain_key_get_iteration
	sender_chain_key_get_seed
	sender_key_distribution_message_copy
	sender_key_distribution_message_create
	sender_key_distribution_message_deserialize
	sender_key_distribution_message_destroy
	sender_key_distribution_message_get_chain_key
	sender_key_distribution_message_get_id
	sender_key_distribution_message_get_iteration
	sender_key_distribution_message_get_signature_key
	sender_key_message_copy
	sender_key_message_create
	sender_key_message_deserialize
	sender_key_message_destroy
	sender_key_message_get_ciphertext
	sender_key_message_get_iteration
	sender_key_message_get_key_id
	sender_key_message_verify_signature
	sender_key_record_add_sender_key_state
	sender_key_record_copy
	sender_key_record_create
	sender_key_record_deserialize
	sender_key_record_destroy
	sender_key_record_get_sender_key_state
	sender_key_record_get_sender_key_state_by_id
	sender_key_record_get_user_record
	sender_key_record_is_empty
	sender_key_record_serialize
	sender_key_record_set_sender_key_state
	sender_key_record_set_user_record
	sender_key_state_add_sender_message_key
	sender_key_state_copy
	sender_key_state_create
	sender_key_state_deserialize
	sender_key_state_destroy
	sender_key_state_get_chain_key
	sender_key_state_get_key_id
	sender_key_state_get_signing_key_private
	sender_key_state_get_signing_key_public
	sender_key_state_has_sender_message_key
	sender_key_state_remove_sender_message_key
	sender_key_state_serialize
	sender_key_state_set_chain_key
	sender_message_key_create
	sender_message_key_destroy
	sender_message_key_get_cipher_key
	sender_message_key_get_iteration
	sender_message_key_get_iv
	sender_message_key_get_seed
	session_builder_create
	session_builder_free
	session_builder_get_version
	session_builder_process_pre_key_bundle
	session_builder_set_version
	session_cipher_create
	session_cipher_decrypt_pre_key_signal_message
	session_cipher_decrypt_signal_message
	session_cipher_encrypt
	session_cipher_free
	session_cipher_get_remote_registration_id
	session_cipher_get_session_version
	session_cipher_get_user_data
	session_cipher_get_version
	session_cipher_set_decryption_callback
	session_cipher_set_user_data
	session_cipher_set_version
	session_pre_key_bundle_create
	session_pre_key_bundle_destroy
	session_pre_key_bundle_get_device_id
	session_pre_key_bundle_get_identity_key
	session_pre_key_bundle_get_pre_key
	session_pre_key_bundle_get_pre_key_id
	session_pre_key_bundle_get_registration_id
	session_pre_key_bundle_get_signed_pre_key
	session_pre_key_bundle_get_signed_pre_key_id
	session_pre_key_bundle_get_signed_pre_key_signature
	session_pre_key_create
	session_pre_key_deserialize
	session_pre_key_destroy
	session_pre_key_get_id
	session_pre_key_get_key_pair
	session_pre_key_serialize
	session_record_archive_current_state
	session_record_copy
	session_record_create
	session_record_deserialize
	session_record_destroy
	session_record_get_previous_states_element
	session_record_get_previous_states_head
	session_record_get_previous_states_next
	session_record_get_previous_states_remove
	session_record_get_state
	session_record_get_user_record
	session_record_get_version
	session_record_has_session_state
	session_record_is_fresh
	session_record_promote_state
	session_record_serialize
	session_record_set_state
	session_record_set_user_record
	session_record_set_version
	session_signed_pre_key_create
	session_signed_pre_key_deserialize
	session_signed_pre_key_destroy
	session_signed_pre_key_get_id
	session_signed_pre_key_get_key_pair
	session_signed_pre_key_get_signature
	session_signed_pre_key_get_signature_len
	session_signed_pre_key_get_signature_omemo
	session_signed_pre_key_get_signature_omemo_len
	session_signed_pre_key_get_timestamp
	session_signed_pre_key_serialize
	session_state_add_receiver_chain
	session_state_clear_unacknowledged_pre_key_message
	session_state_copy
	session_state_create
	session_state_deserialize
	session_state_destroy
	session_state_get_alice_base_key
	session_state_get_kdf_infos
	session_state_get_local_identity_key
	session_state_get_local_is_alice
	session_state_get_local_registration_id
	session_state_get_needs_refresh
	session_state_get_pending_key_exchange_base_key
	session_state_get_pending_key_exchange_identity_key
	session_state_get_pending_key_exchange_ratchet_key
	session_state_get_pending_key_exchange_sequence
	session_state_get_previous_counter
	session_state_get_receiver_chain_key
	session_state_get_remote_identity_key
	session_state_get_remote_registration_id
	session_state_get_root_key
	session_state_get_sender_chain_key
	session_state_get_sender_ratchet_key
	session_state_get_sender_ratchet_key_pair
	session_state_get_session_version
	session_state_has_message_keys
	session_state_has_pending_key_exchange
	session_state_has_sender_chain
	session_state_has_unacknowledged_pre_key_message
	session_state_remove_message_keys
	session_state_serialize
	session_state_set_alice_base_key
	session_state_set_local_identity_key
	session_state_set_local_is_alice
	session_state_set_local_registration_id
	session_state_set_message_keys
	session_state_set_needs_refresh
	session_state_set_pending_key_exchange
	session_state_set_previous_counter
	session_state_set_receiver_chain_key
	session_state_set_remote_identity_key
	session_state_set_remote_registration_id
	session_state_set_root_key
	session_state_set_sender_chain
	session_state_set_sender_chain_key
	session_state_set_session_version
	session_state_set_unacknowledged_pre_key_message
	session_state_unacknowledged_pre_key_message_get_base_key
	session_state_unacknowledged_pre_key_message_get_pre_key_id
	session_state_unacknowledged_pre_key_message_get_signed_pre_key_id
	session_state_unacknowledged_pre_key_message_has_pre_key_id
	signal_buffer_alloc
	signal_buffer_append
	signal_buffer_bzero_free
	signal_buffer_compare
	signal_buffer_const_data
	signal_buffer_copy
	signal_buffer_create
	signal_buffer_data
	signal_buffer_free
	signal_buffer_len
	signal_buffer_list_alloc
	signal_buffer_list_at
	signal_buffer_list_bzero_free
	signal_buffer_list_copy
	signal_buffer_list_free
	signal_buffer_list_push_back
	signal_buffer_list_size
	signal_buffer_n_copy
	signal_context_create
	signal_context_destroy
	signal_context_set_crypto_provider
	signal_context_set_locking_functions
	signal_context_set_log_function
	signal_int_list_alloc
	signal_int_list_at
	signal_int_list_free
	signal_int_list_push_back
	signal_int_list_size
	signal_message_copy
	signal_message_create
	signal_message_deserialize
	signal_message_deserialize_omemo
	signal_message_destroy
	signal_message_get_body
	signal_message_get_counter
	signal_message_get_message_version
	signal_message_get_sender_ratchet_key
	signal_message_is_legacy
	signal_message_verify_mac
	signal_protocol_identity_get_key_pair
	signal_protocol_identity_get_local_registration_id
	signal_protocol_identity_is_trusted_identity
	signal_protocol_identity_save_identity
	signal_protocol_key_helper_generate_identity_key_pair
	signal_protocol_key_helper_generate_pre_keys
	signal_protocol_key_helper_generate_registration_id
	signal_protocol_key_helper_generate_sender_key
	signal_protocol_key_helper_generate_sender_key_id
	signal_protocol_key_helper_generate_sender_signing_key
	signal_protocol_key_helper_generate_signed_pre_key
	signal_protocol_key_helper_get_random_sequence
	signal_protocol_key_helper_key_list_element
	signal_protocol_key_helper_key_list_free
	signal_protocol_key_helper_key_list_next
	signal_protocol_key_helper_upgrade_signed_pre_key
	signal_protocol_pre_key_contains_key
	signal_protocol_pre_key_load_key
	signal_protocol_pre_key_remove_key
	signal_protocol_pre_key_store_key
	signal_protocol_sender_key_load_key
	signal_protocol_sender_key_store_key
	signal_protocol_session_contains_session
	signal_protocol_session_delete_all_sessions
	signal_protocol_session_delete_session
	signal_protocol_session_get_sub_device_sessions
	signal_protocol_session_load_session
	signal_protocol_session_store_session
	signal_protocol_signed_pre_key_contains_key
	signal_protocol_signed_pre_key_load_key
	signal_protocol_signed_pre_key_remove_key
	signal_protocol_signed_pre_key_store_key
	signal_protocol_store_context_create
	signal_protocol_store_context_destroy
	signal_protocol_store_context_set_identity_key_store
	signal_protocol_store_context_set_pre_key_store
	signal_protocol_store_context_set_sender_key_store
	signal_protocol_store_context_set_session_store
	signal_protocol_store_context_set_signed_pre_key_store
	signal_type_ref
	signal_type_unref
	symmetric_signal_protocol_parameters_create
	symmetric_signal_protocol_parameters_destroy
	symmetric_signal_protocol_parameters_get_our_base_key
	symmetric_signal_protocol_parameters_get_our_identity_key
	symmetric_signal_protocol_parameters_get_our_ratchet_key
	symmetric_signal_protocol_parameters_get_their_base_key
	symmetric_signal_protocol_parameters_get_their_identity_key
	symmetric_signal_protocol_parameters_get_their_ratchet_key
